var __extends = (this && this.__extends) || (function () {
    var extendStatics = Object.setPrototypeOf ||
        ({ __proto__: [] } instanceof Array && function (d, b) { d.__proto__ = b; }) ||
        function (d, b) { for (var p in b) if (b.hasOwnProperty(p)) d[p] = b[p]; };
    return function (d, b) {
        extendStatics(d, b);
        function __() { this.constructor = d; }
        d.prototype = b === null ? Object.create(b) : (__.prototype = b.prototype, new __());
    };
})();
var __assign = (this && this.__assign) || Object.assign || function(t) {
    for (var s, i = 1, n = arguments.length; i < n; i++) {
        s = arguments[i];
        for (var p in s) if (Object.prototype.hasOwnProperty.call(s, p))
            t[p] = s[p];
    }
    return t;
};
import * as React from "react";
import { diff } from "./utils";
var Scheduler = (function (_super) {
    __extends(Scheduler, _super);
    function Scheduler() {
        var _this = _super !== null && _super.apply(this, arguments) || this;
        _this.queue = [];
        _this.timeout = 0;
        _this.sortOnQ = false;
        _this.scheduledCallbacks = new Map();
        _this.state = {
            recalculate: 0,
            queue: _this.queue,
            current: { ref: null }
        };
        _this.add = function (cb, _a, ref) {
            var _b = _a.priority, priority = _b === void 0 ? 0xFFFFFF : _b, _c = _a.shift, shift = _c === void 0 ? 0 : _c;
            var q = {
                cb: cb,
                priority: priority,
                shift: shift,
                index: _this.queue.length,
                sortIndex: 0,
                sortOrder: priority,
                ref: ref,
                executed: false
            };
            q.sortIndex = q.index + q.shift;
            _this.queue.push(q);
            _this.update();
            return q;
        };
        _this.remove = function (q) {
            var index = _this.queue.indexOf(q);
            if (index > 0) {
                _this.queue.splice(index, 1);
            }
        };
        _this.replace = function (q, cb, _a, ref) {
            var _b = _a.priority, priority = _b === void 0 ? 0 : _b, _c = _a.shift, shift = _c === void 0 ? 0 : _c;
            if (q) {
                var changedPriority = q.priority !== priority || q.shift !== shift;
                var changedRef = q.ref !== ref;
                q.cb = cb;
                q.priority = priority;
                q.shift = shift;
                q.sortIndex = q.index + q.shift;
                q.ref = ref;
                if (changedPriority || changedRef) {
                    _this.update();
                }
            }
            return q;
        };
        _this.reset = function () {
            _this.queue.forEach(function (q) { return q.executed = false; });
            _this.update();
        };
        _this.schedule = function (ref, cb) {
            _this.scheduledCallbacks.set(ref, cb);
            _this.update();
        };
        _this.channel = {
            add: _this.add,
            remove: _this.remove,
            replace: _this.replace,
            reset: _this.reset,
            schedule: _this.schedule
        };
        return _this;
    }
    Scheduler.prototype.componentWillUnmount = function () {
        window.clearTimeout(this.timeout);
    };
    Scheduler.prototype.componentDidUpdate = function (props) {
        if (this.props.observe !== props.observe ||
            this.props.reverse !== props.reverse ||
            this.props.disabled !== props.disabled) {
            this.update();
        }
    };
    Scheduler.prototype.update = function () {
        this.sortOnQ = true;
        this.scheduleQ(this.props.noInitialDelay && this.noExecutedTask()
            ? 0
            : this.props.stepDelay);
    };
    Scheduler.prototype.scheduleQ = function (when) {
        var _this = this;
        if (!this.timeout && !this.props.disabled) {
            this.timeout = window.setTimeout(function () {
                _this.timeout = 0;
                var cbs = _this.scheduledCallbacks;
                _this.scheduledCallbacks = new Map();
                cbs.forEach(function (cb) { return cb(); });
                if (_this.sortOnQ) {
                    _this.sortOnQ = false;
                    _this.sortQ();
                }
                _this.executeQ();
            }, when);
        }
    };
    Scheduler.prototype.sortQ = function () {
        var _this = this;
        if (this.props.source) {
            this.queue.forEach(function (q) { return q.sortOrder = q.priority < Infinity ? _this.props.source(q) : q.priority; });
        }
        else {
            this.queue.forEach(function (q) { return q.sortOrder = q.priority; });
        }
        if (this.props.reverse) {
            this.queue.sort(function (a, b) { return diff(b, a); });
        }
        else {
            this.queue.sort(function (a, b) { return diff(a, b); });
        }
    };
    Scheduler.prototype.nextQ = function () {
        var left = this.queue.filter(function (_a) {
            var executed = _a.executed, priority = _a.priority, sortOrder = _a.sortOrder;
            return !executed && sortOrder < Infinity && priority < Infinity;
        });
        return left[0];
    };
    Scheduler.prototype.executeQ = function () {
        var _this = this;
        var q = this.nextQ();
        this.state.current.ref = __assign({}, q);
        if (q) {
            q.executed = true;
            Promise.resolve(q.cb())
                .then(function (result) {
                var dt = (typeof result === "number" ? result : 0) || 0;
                var when = Math.max(0, _this.props.stepDelay + dt);
                _this.scheduleQ(when);
                if (_this.props.withSideEffect) {
                    _this.setState({
                        recalculate: 1
                    });
                }
            });
        }
        else {
            this.props.onEmptyQueue && this.props.onEmptyQueue();
        }
    };
    Scheduler.prototype.noExecutedTask = function () {
        return !this.queue.find(function (x) { return x.executed; });
    };
    Scheduler.prototype.render = function () {
        return this.props.children(this.channel);
    };
    return Scheduler;
}(React.Component));
export { Scheduler };
